﻿using SharpDX.DirectInput;
using System;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;

namespace TestSharpDx
{
	// Device container. Inherit from ListViewItem so we can directly add to ListView
	public class DeviceInfo : ListViewItem
	{
		private static DirectInput directInput;
		public static DirectInput DirectInput { get { return directInput; } set { directInput = value; } }

		// Static event goes for all device changes
		private static event EventHandler<DeviceChangedEventArgs> onDeviceChanged;
		public static event EventHandler<DeviceChangedEventArgs> OnDeviceChanged { add { onDeviceChanged += value; } remove { onDeviceChanged -= value; } }

		// When joystick position changes
		private event EventHandler<PositionChangedEventArgs> onPositionChanged;
		public event EventHandler<PositionChangedEventArgs> OnPositionChanged { add { onPositionChanged += value; } remove { onPositionChanged -= value; } }

		// some button pressed
		private event EventHandler<SwitchChangedEventArgs> onSwitchChanged;
		public event EventHandler<SwitchChangedEventArgs> OnSwitchChanged { add { onSwitchChanged += value; } remove { onSwitchChanged -= value; } }

		private Int32 x = 0;
		public Int32 X 
		{
			get 
			{ return x; } 
			set 
			{
				if (x != value) // Only fire an event when something changes
				{
					EventHandler<PositionChangedEventArgs> handler = onPositionChanged;
					if (handler != null)
						handler(this, new PositionChangedEventArgs(this, JoystickOffset.X, value));

					x = value;
				}
			}
		}
		private Int32 y = 0;
		public Int32 Y
		{
			get
			{ return y; }
			set
			{
				if (y != value)// Only fire an event when something changes
				{
					EventHandler<PositionChangedEventArgs> handler = onPositionChanged;
					if (handler != null)
						handler(this, new PositionChangedEventArgs(this, JoystickOffset.Y, value));

					y = value;
				}
			}
		}

		private bool usable;
		public bool Usable { get { return usable; } }
		private Joystick joystick;
		public Joystick Joystick { get { return joystick; } }
		private Guid deviceGuid;
		public Guid DeviceGuid { get { return deviceGuid; } }
		private DeviceType typeDevice;
		public DeviceType TypeDevice { get { return typeDevice; } }
		private EffectInfo[] allEffects;
		public EffectInfo[] AllEffects { get { return allEffects; } set { allEffects = value; } }

		public int ButtonCount {get{return ((joystick != null)?joystick.Capabilities.ButtonCount:0);}}

		private bool[] buttons;
		public bool[] Buttons { get { return buttons; } }

		// constructor
		public DeviceInfo(DeviceType _typeDevice, Guid _deviceGuid)
		{
			deviceGuid = _deviceGuid;
			typeDevice = _typeDevice;
			joystick = new Joystick(directInput, _deviceGuid);

			// Set BufferSize in order to use buffered data. (do this before Aquire())
			joystick.Properties.BufferSize = 128;

			usable = false;
			Name = "";
			try
			{
				// Acquire the joystick
				joystick.Acquire();
				Name = joystick.Properties.ProductName;
				usable = true;
			}
			catch (SharpDX.SharpDXException)
			{
				Name = "Not implemented";
			}
			catch (System.Exception)
			{
				Name = "Unknown";
			}

			Debug.WriteLine(String.Format("* {0} GUID: {1} >> {2} {3}", typeDevice, deviceGuid, Name, (usable ? "Usable" : "Not supported")));

			this.Text = Name;
			this.SubItems.Add(typeDevice.ToString());
			this.SubItems.Add(deviceGuid.ToString());
			this.BackColor = usable ? SystemColors.Control : SystemColors.ControlDark;

			if (usable)
			{
				// Query all supported ForceFeedback effects
				EffectInfo[] allEffects = joystick.GetEffects().ToArray();

				buttons = new bool[joystick.Capabilities.ButtonCount];

				foreach (var effectInfo in allEffects)
				{ Console.WriteLine("Effect available {0}", effectInfo.Name); }
			}
		}

		public void Start()
		{
			joystick.Acquire();
			EventHandler<DeviceChangedEventArgs> handler = onDeviceChanged;
			if (handler != null)
				handler(this, new DeviceChangedEventArgs(this, true));
		}
		public void Stop()
		{
			EventHandler<DeviceChangedEventArgs> handler = onDeviceChanged;
			if (handler != null)
				handler(this, new DeviceChangedEventArgs(this, false));
			joystick.Unacquire();
		}

		public void Poll()
		{
			JoystickUpdate[] allData = joystick.GetBufferedData();
			foreach (var data in allData)
			{
				if (data.Offset == JoystickOffset.X)
				{
					X = data.Value;
				}
				else if (data.Offset == JoystickOffset.Y)
				{
					Y = data.Value;
				}
				else if (data.Offset == JoystickOffset.Z)
				{ }
				else if (data.Offset >= JoystickOffset.Buttons0 && data.Offset <= JoystickOffset.Buttons127)
				{
					Debug.WriteLine(String.Format("Button: {0} = {1}", data.Offset, data.Value));
					// seems at most joysticks the button value is 0 (off) or 127(pushed)

					int btn = data.Offset - JoystickOffset.Buttons0; // btn index 0 - ...
					buttons[btn] = (data.Value != 0);

					EventHandler<SwitchChangedEventArgs> handler = onSwitchChanged;
					if (handler != null)
						handler(this, new SwitchChangedEventArgs(this, btn, buttons[btn]));
				}
			}
		}
	}
}
