﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using SharpDX;
using SharpDX.DirectInput;
using System.Diagnostics;

namespace TestSharpDx
{
	public partial class FrmTestSharpDx : Form
	{
		// the hardware connection
		private DirectInput directInput = null;

		// The current device
		private DeviceInfo activeDevice = null;

		public FrmTestSharpDx()
		{
			DoubleBuffered = true; // avoid flickering display
			InitializeComponent();
		}

		private void FormTestSharpDx_Load(object sender, EventArgs e)
		{
			// joystick output is 0 to Uint16.MaxValue (65535)
			// See remarks
			barX.Minimum = 0;
			barX.Maximum = UInt16.MaxValue;
			barY.Minimum = 0;
			barY.Maximum = UInt16.MaxValue;

			// Init the ListView by adding some columns
			listDevices.View = View.Details;
			listDevices.MultiSelect = false;
			listDevices.HideSelection = false;
			listDevices.Columns.Add("Name", 150, HorizontalAlignment.Left);
			listDevices.Columns.Add("Type", 100, HorizontalAlignment.Left);
			listDevices.Columns.Add("GUID", -2, HorizontalAlignment.Left);

			DeviceInfo.OnDeviceChanged += DeviceInfo_OnDeviceChanged;
			ReadControllers();
		}

		void DeviceInfo_OnDeviceChanged(object sender, DeviceChangedEventArgs e)
		{
			panelButtons.Invalidate();
			panelPosition.Invalidate();
		}

		// get data for connected joysticks
		private void ReadControllers()
		{
			// Initialize DirectInput
			directInput = new DirectInput();
			DeviceInfo.DirectInput = directInput;

			// check all kind of controllers: joysticks, pads, keyboards etc
			foreach (DeviceType dt in Enum.GetValues(typeof(DeviceType)))
			{
				foreach (var deviceInstance in directInput.GetDevices(dt, DeviceEnumerationFlags.AllDevices))
				{
					// Create a container
					DeviceInfo di = new DeviceInfo(dt, deviceInstance.InstanceGuid);

					// Put this container in a list view
					listDevices.Items.Add(di);
				}
			}

			if (listDevices.Items.Count == 0)
				txtName.Text = "No joystick/Gamepad/controller found";
			else
			{
				foreach (DeviceInfo di in listDevices.Items)
				{
					if (activeDevice == null && di.Usable && di.TypeDevice == DeviceType.Joystick)
					{
						di.Selected = true;
						break;
					}
				}
			}
		}

		private void timerDisplay_Tick(object sender, EventArgs e)
		{
			if (activeDevice == null)
				return;

			// In a bigger program this poll is called from a thread. now we just use the form timer
			// threading would introduce some cross-threading problems (use Invoke to avoid that)
			activeDevice.Poll();
		}

		private int radius = 4;
		private void panelPosition_Paint(object sender, PaintEventArgs e)
		{
			if (activeDevice == null)
			{
				e.Graphics.DrawString("no device", this.Font, Brushes.Black, ClientRectangle);
				return;
			}

			System.Drawing.Point pt = new System.Drawing.Point(
				(int)Math.Round((panelPosition.Width * activeDevice.X / (double)UInt16.MaxValue) - radius),
				(int)Math.Round((panelPosition.Height * activeDevice.Y / (double)UInt16.MaxValue) - radius)
				);

			//Debug.WriteLine(String.Format("({0};{1})  ({2};{3})", activeDevice.X, activeDevice.Y, pt.X, pt.Y));

			try
			{
				e.Graphics.DrawLine(Pens.Black, new System.Drawing.Point(panelPosition.Width / 2, 0), new System.Drawing.Point(panelPosition.Width / 2, panelPosition.Height));
				e.Graphics.DrawLine(Pens.Black, new System.Drawing.Point(0, panelPosition.Height / 2), new System.Drawing.Point(panelPosition.Width, panelPosition.Height / 2));
				e.Graphics.DrawEllipse(Pens.Red, new System.Drawing.Rectangle(pt, new Size(radius * 2, radius * 2)));
			}
			catch (System.Exception ex)
			{
				Debug.WriteLine("Something wrong with drawing: " + ex.Message);
			}

			try
			{
				barX.Value = Global.Clamp(activeDevice.X,  barX.Minimum, barX.Maximum);
				barY.Value = Global.Clamp(activeDevice.Y,  barY.Minimum, barY.Maximum);
				lblJoystickX.Text = activeDevice.X.ToString();
				lblJoystickY.Text = activeDevice.Y.ToString();
			}
			catch (System.Exception ex)
			{
				Debug.WriteLine("Something wrong with bars: " + ex.Message);
			}
		}

		bool busy = false;
		private void listDevices_SelectedIndexChanged(object sender, EventArgs e)
		{
			 if (busy || listDevices.SelectedItems.Count == 0) return;
			busy = true;

			// the selected item
			DeviceInfo sel = (DeviceInfo)listDevices.SelectedItems[0];

			if (!sel.Usable) // just set selection back to activeDevice
			{
				Debug.WriteLine("set back old 1");
				listDevices.SelectedIndices.Clear();

				if (activeDevice != null)
					activeDevice.Selected = true; // this will make activeDevice come back here in this function! bool busy avoids unnecessarily work
			}
			else if (sel.Usable && sel.Selected && sel != activeDevice)
			{
				// new selected device
				if (activeDevice != null)
				{
					Debug.Write("stop ");
					activeDevice.OnPositionChanged -= activeDevice_OnPositionChanged;
					activeDevice.OnSwitchChanged -= activeDevice_OnSwitchChanged;
					activeDevice.Stop();
				}

				Debug.Write("start ");

				activeDevice = sel;
				activeDevice.Start();

				activeDevice.OnPositionChanged += activeDevice_OnPositionChanged;
				activeDevice.OnSwitchChanged += activeDevice_OnSwitchChanged;

				txtName.Text = activeDevice.Joystick.Properties.ProductName;
				activeDevice.X = 0;
				activeDevice.Y = 0;
				Debug.WriteLine("set");

			}
			else if (sel.Usable && sel.Selected && sel == activeDevice)
			{
				Debug.WriteLine("same sel");
				//activeDevice.Focused = true;
				//listDevices.Select();
			}

			busy = false;

/*
			// Retrieve range did not work. not implemented in SharpDx
			try
			{
				DeviceProperties pr = activeJoystick.Properties;
				Debug.WriteLine(String.Format("Min: {0} Max:{1}",
				activeJoystick.Properties.Range.Minimum,
				activeJoystick.Properties.Range.Maximum));
			}
			catch (System.Exception ex)
			{
				Debug.WriteLine("boem: " + ex.Message);
			}
*/
		}

		void activeDevice_OnSwitchChanged(object sender, SwitchChangedEventArgs e)
		{
			panelButtons.Invalidate();
		}

		void activeDevice_OnPositionChanged(object sender, PositionChangedEventArgs e)
		{
			panelPosition.Invalidate();
		}

		private int btnSize = 6;
		private void panelButtons_Paint(object sender, PaintEventArgs e)
		{
			if (activeDevice == null)
			{
				e.Graphics.DrawString("no device", this.Font, Brushes.Black, ClientRectangle);
				return;
			}
			if (activeDevice.ButtonCount == 0)
			{
				e.Graphics.DrawString("no buttons", this.Font, Brushes.Black, ClientRectangle);
				return;
			}
			int yPos = 10;
			for (int i = 0; i < activeDevice.Buttons.Length; i++)
			{
				e.Graphics.DrawString("Btn" + i.ToString(), this.Font, Brushes.Black, new System.Drawing.Point(25, yPos));

				System.Drawing.Point pt = new System.Drawing.Point(5, yPos);

				e.Graphics.FillEllipse(
					activeDevice.Buttons[i]?Brushes.Red:Brushes.Green, new System.Drawing.Rectangle(pt, new Size(btnSize * 2, btnSize * 2)));

				yPos += 15;
			}
		}

		private void FrmTestSharpDx_Resize(object sender, EventArgs e)
		{
			panelButtons.Invalidate();
			panelPosition.Invalidate();
		}
	}
}
